# ***************************************************************************
# Copyright 2019 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************

from cli.commands.cli_command_provider import CliCommandProvider
from cli.utils.constants import NODE_TYPE_EDGE
from cli.utils.constants import NODE_TYPE_AUTONOMOUS_EDGE
from cli.utils.constants import NODE_TYPE_HOST_CENTOSKVM
from cli.utils.constants import NODE_TYPE_HOST_SLESKVM
from cli.utils.constants import NODE_TYPE_HOST_ESX
from cli.utils.constants import NODE_TYPE_MANAGER
from cli.utils.constants import NODE_TYPE_CONTROLLER
from cli.utils.constants import NODE_TYPE_HOST_RHELKVM
from cli.utils.constants import NODE_TYPE_HOST_UBUNTUKVM
from cli.utils.constants import NODE_TYPE_BAREMETAL_UBUNTU_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_SLES_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_CONTAINER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_CONTAINER
from cli.utils.error_code import ErrorCode as ec
from cli.utils.node_utils import get_node_type
from cli.utils.signal_utils import send_sighup

import errno
import json
from mako import compat
import nsx_logging
import os

lg = nsx_logging.getLogger("cli.commands.common.log_level_config_utils")

LOG_LEVEL_MAP = {
    "off": "0",
    "fatal": "1",
    "error": "2",
    "warn": "3",
    "info": "4",
    "debug": "5",
    "trace": "6"
}

# Use nsx-manager if exists, else use the first node type
node_types = get_node_type().split()
if NODE_TYPE_MANAGER in node_types:
    g_node_type = NODE_TYPE_MANAGER
else:
    g_node_type = node_types[0]


def update_config_file(daemon_name, level):
    """Helper method to set the log level in /etc/vmware/<service-name>/config.json file
       Param : daemon_name - The name of the service you want to set the configuration for
       Param : level - The loglevel you want to set for the service (range[0,6])
       Returns True if successful in setting the log level to the file. Else returns False
    """
    filename = "/etc/vmware/" + daemon_name + "/config.json"
    logLevel = LOG_LEVEL_MAP[level]
    if os.path.isfile(filename):
        with open(filename, 'r+') as f:
            data = json.load(f)
            data['logLevel'] = logLevel
            f.seek(0)
            json.dump(data, f, indent=4)
    else:
        if g_node_type in [NODE_TYPE_EDGE, NODE_TYPE_AUTONOMOUS_EDGE, \
            NODE_TYPE_CONTROLLER, NODE_TYPE_MANAGER]:
            lg.error("Config file %s not found", filename, error_code=ec.OPERATION_FAILED)
            return False

        try:
            original_umask = os.umask(0)
            os.makedirs(os.path.dirname(filename), compat.octal("0775"))
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                lg.error("Could not create config dir %s exception %s", \
                    os.path.dirname(filename), exc, error_code=ec.OPERATION_FAILED)
                return False
        finally:
            os.umask(original_umask)
        with open(filename, 'w') as f:
            data = {}
            data["logLevel"] = logLevel
            json.dump(data, f, indent=4)

    return True


def read_config_file(daemon_name):
    """Helper method to get the log level from /etc/vmware/<service-name>/config.json file
       Param : daemon_name - The name of the service you want to set the configuration for
       Returns:
          logLevel if successful in getting the log level from the file
             One of: "off", "fatal", "error", "info", "debug", "trace"
          else returns False
    """
    GET_LOG_LEVEL_MAP = {value: key for key, value in list(LOG_LEVEL_MAP.items())}
    filename = "/etc/vmware/" + daemon_name + "/config.json"
    logLevel = False
    if os.path.isfile(filename):
        try:
            with open(filename, 'r') as f:
                data = json.load(f)
                level = data['logLevel']
                logLevel = GET_LOG_LEVEL_MAP[level]
        except:
            lg.error("Could not read the config file for %s" % daemon_name)
            return False
    else:
        lg.error("Config file for %s is not present" % daemon_name)
        return False
    return logLevel
