# ***************************************************************************
# Copyright 2018 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import json
import nsx_logging
import sys
import uuid

from cli.commands.cli_command_provider import CliCommandProvider
from cli.utils.nestdb_utils import invoke_nestdb_command, query_type
from cli.utils.ipaddr_utils import mac_int_to_string, ip_int_to_string

LOG_SWITCH_CONFIG_MSG = "vmware.nsx.nestdb.LogSwitchConfigMsg"
ROUTING_DOMAIN_VTEP_MSG = "vmware.nsx.nestdb.RoutingDomainVtepMsg"
ROUTING_DOMAIN_FIB_MSG = "vmware.nsx.nestdb.RoutingDomainFibMsg"

lg = nsx_logging.getLogger("cli.commands.lcp_common.nsx_proxy.routing_domain")


class NestDBL2Convertor():
    @staticmethod
    def vtep_cache_convertor(vtep):
        entry = {}
        entry["vtep_ip"] = ip_int_to_string(vtep["vtep_ip"]["ipv4"])
        entry["subnet_ip"] = ip_int_to_string(vtep["segment_id"]["ipv4"])
        entry["label"] = vtep["vtep_label"]["label"]
        entry["vtep_mac"] = mac_int_to_string(vtep["vtep_mac"]["mac"])
        return entry

    @staticmethod
    def get_routing_domain_msg(nestdb_schema, rd):
        rd_uuid_int = uuid.UUID(rd).int
        rd_uuid_int_right = rd_uuid_int & (1<<64) - 1
        rd_uuid_int_left = rd_uuid_int >> 64
        rd_uuid = "{\"right\": " + str(rd_uuid_int_right) + ", \"left\":" + str(rd_uuid_int_left) + "}"
        cmd = ["query %s {\"id\":{\"self\":{\"op\":\"EQUALS\"}}} {\"id\":%s}"
               % (nestdb_schema, rd_uuid)]
        return invoke_nestdb_command(cmd)

    @staticmethod
    def get_valid_routing_domain_uuid(log_switch_config):
        left = log_switch_config["value"]["routing_domain_id"]["left"]
        right = log_switch_config["value"]["routing_domain_id"]["right"]

        if left == 0 and right == 0:
            return None
        else:
            rd_uuid_int = uuid.UUID(int = left << 64 | right)
            return str(rd_uuid_int)

    @staticmethod
    def get_routing_domain_ids(nestdb_schema):
        cmd_list = ["get", nestdb_schema]
        rc, rv, err = invoke_nestdb_command(cmd_list)
        if rc != 0:
            return (rc, rv, err)

        if rv is None:
            return (0, None, "not exist")

        result = []
        ls_configs = json.loads(rv)
        if "result" in ls_configs:
            for item in ls_configs["result"]:
                rd_uuid = NestDBL2Convertor.get_valid_routing_domain_uuid(item)
                if rd_uuid is not None:
                    result.append(rd_uuid)
        return (0, list(set(result)), None)

    @staticmethod
    def parse_rd_msg(nestdb_schema, rd):
        rc, rv, err = NestDBL2Convertor.get_routing_domain_msg(nestdb_schema, rd)
        if rc != 0:
            return (rc, rv, err)

        if rv is None:
            return (0, None, "not exist")

        resp = json.loads(rv)
        if "result" in resp:
            out = [v["value"] for v in resp["result"]]
            return (0, out, None)
        else:
            return (0, None, "not exist")

    @staticmethod
    def get_rd_cache_by_key(nestdb_schema, rd, key):
        _L2_CONVERTOR_MAPPING = {
            'vtep': NestDBL2Convertor.vtep_cache_convertor,
        }

        rc, rv, err = NestDBL2Convertor.parse_rd_msg(nestdb_schema, rd)
        if rc != 0:
            return None

        entry_list = []
        if rv is None:
            return json.dumps(entry_list)

        for val in rv:
            if key in val:
                for entry in val[key]:
                    entry_list.append(_L2_CONVERTOR_MAPPING[key](entry))

        return entry_list


class GetRoutingDomainLocalVTEP(CliCommandProvider):
    def __call__(self, cmd_list):
        rd = cmd_list[4]
        local_vtep = NestDBL2Convertor.get_rd_cache_by_key(ROUTING_DOMAIN_VTEP_MSG,
                                                           rd, 'vtep')
        if local_vtep is None:
            return "MSG_NSX_PROXY_ROUTING_DOMAIN_TABLE_FAILED"
        return "OK\n\n%s" % json.dumps(local_vtep)


class GetRoutingDomainRemoteVTEP(CliCommandProvider):
    def __call__(self, cmd_list):
        rd = cmd_list[4]
        remote_vtep = NestDBL2Convertor.get_rd_cache_by_key(ROUTING_DOMAIN_FIB_MSG,
                                                            rd, 'vtep')
        if remote_vtep is None:
            return "MSG_NSX_PROXY_ROUTING_DOMAIN_TABLE_FAILED"

        local_vtep = NestDBL2Convertor.get_rd_cache_by_key(ROUTING_DOMAIN_VTEP_MSG,
                                                           rd, 'vtep')
        if local_vtep is None:
            return "MSG_NSX_PROXY_ROUTING_DOMAIN_TABLE_FAILED"

        for elem in local_vtep:
            if elem in remote_vtep:
                remote_vtep.remove(elem)

        return "OK\n\n%s" % json.dumps(remote_vtep)

def get_routing_domain_ids():
    rc, rds, err = NestDBL2Convertor.get_routing_domain_ids(LOG_SWITCH_CONFIG_MSG)
    if rc != 0:
        lg.error("MSG_NSX_PROXY_ROUTING_DOMAIN_IDS_FAILED")
        return "MSG_NSX_PROXY_ROUTING_DOMAIN_IDS_FAILED"
    else:
        return rds
