# **********************************************************************
# Copyright 2017 VMware, Inc.  All rights reserved. VMware Confidential.
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
import os
import tempfile
from subprocess import PIPE
from subprocess import Popen

import nsx_logging
from cli.utils.command_utils import invoke_command
from cli.utils.location_utils import FilePaths
from cli.utils.node_utils import is_windows

lg = nsx_logging.getLogger("cli.utils.cert_utils")

_CMD_OPENSSL = FilePaths.OPENSSL_PATH
_SIG_SHA256_FINGERPRINT = "SHA256 Fingerprint="
_SIG_SUBJECT_NAME = "subject="
_BEGIN_CERT_TOKEN = "-----BEGIN CERTIFICATE-----"


def get_cert_fingerprint(cert_file):
    """Get sha256 fingerprint of specified cert_file.

    Args:
        cert_file: Path to .pem certificate file, caller must have read perms.
    Return:
        Lower case sha256 fingerprint with all colons removed, or None if error.
    """
    cmd = [
        _CMD_OPENSSL, "x509", "-fingerprint", "-noout", "-sha256",
        "-in", cert_file]
    rc, out, err = invoke_command(cmd)
    if rc == 0 and out.startswith(_SIG_SHA256_FINGERPRINT):
        parts = out.split("=", 1)
        return parts[-1].replace(":", "").lower().strip()
    lg.error(
        "Error getting fingerprint of cert file: %s, rc: %s, out: %s, err: %s",
        cert_file, rc, out, err)
    return None


def create_self_signed_cert(openssl_cnf, key_file, cert_file, key_type, days):
    """Create key and self signed cert using openssl_cnf.

    Args:
        openssl_cnf: Path to openssl config.
        key_file: Path to .pem key file.
        cert_file: Path to .pem certificate file.
    Return:
        True on success, False on failure.
    """
    cmd = [
        _CMD_OPENSSL, "req", "-new", "-newkey", key_type, "-days", str(days),
        "-nodes", "-x509", "-keyout", key_file, "-out", cert_file, "-config",
        openssl_cnf, "-extensions", "req_ext"]
    rc, out, err = invoke_command(cmd)
    if rc == 0:
        return True
    lg.error(
        "Error generating key and cert files: %s, %s, rc: %s, out: %s, err: %s",
        key_file, cert_file, rc, out, err)
    return False


def get_cert_subject(cert_file):
    """Get 'Subject' from cert_file.

    Args:
        cert_file: Path to .pem certificate file, caller must have read perms.
    Return:
        Subject name if success, None if error.
    """
    cmd = [_CMD_OPENSSL, "x509", "-in", cert_file, "-subject", "-noout"]
    rc, out, err = invoke_command(cmd)
    if rc == 0 and out.startswith(_SIG_SUBJECT_NAME):
        parts = out.split("=", 1)
        return parts[1].strip()
    lg.error(
        "Error getting subject from cert file: %s, rc: %s, out: %s, err: %s",
        cert_file, rc, out, err)
    return None


def get_cert_text_dict(pem_cert_data):
    """Get certificate text.

    Args:
        pem_cert_data: Contents of pem_encoded data.
    Return:
        dict of decoded certificate text data.
    """
    from cryptography import x509
    from cryptography.hazmat.backends import default_backend

    # Get cert data in bytes
    pem_cert_data = pem_cert_data.encode()
    cert = x509.load_pem_x509_certificate(pem_cert_data, default_backend())

    def resolve_rdns(rdns):
        result = []
        for rd in rdns:
            attrs = [i for i in rd._attributes]
            result.append((attrs[0].oid._name, attrs[0].value))
        return result

    _ISSUER_MAP = {
        'countryName': "C",
        'stateOrProvinceName': "ST",
        'localityName': "L",
        'organizationName': "O",
        'organizationalUnitName': "OU",
        'commonName': "CN"
    }

    result = {
        'data': {
            'version': (hex(cert.version.value), cert.version.name),
            'serial_number': hex(cert.serial)
        },
        'signature_info': {
            'algorithm': cert.signature_hash_algorithm.name,
            'issuer': ", ".join(["%s=%s" % (_ISSUER_MAP.get(n, n), v)
                                 for n, v in resolve_rdns(cert.issuer.rdns)]),
            'validity': {
                'not_before':
                    cert.not_valid_before.strftime("%b %d %H:%M:%S %Y") +
                        " GMT",
                'not_after':
                    cert.not_valid_after.strftime("%b %d %H:%M:%S %Y") +
                        " GMT",
            },
            'subject': ", ".join(["%s=%s" % (_ISSUER_MAP.get(n, n), v)
                                  for n, v in resolve_rdns(cert.subject.rdns)]),
            'public_key_info': {
                'public_key_algorithm': cert.public_key().key_size,
                'modulus': hex(cert.public_key().public_numbers().n),
                'exponent': cert.public_key().public_numbers().e
            },
            'signature': ":".join(["{0:0{1}x}".format(i, 2) for i in cert.signature])
        }
    }

    return {'certificate': result}


def generate_certifcate_file(pem):
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as pem_file:
        pem_file.write(pem)
    return pem_file


def remove_certificate_file(file_path):
    if os.path.isfile(file_path):
        try:
            os.unlink(file_path)
        except Exception as ex:
            lg.error("[%s] Unable to remove file: %s", ex, file_path)


def verify_certificate(pem_data):
    """Verifies certificate using openssl.

    Args:
        pem_data: pem certificate string.
    """
    # Ignore human readable part and check only from -----BEGIN CERTIFICATE-----
    pem_data = pem_data[pem_data.index(_BEGIN_CERT_TOKEN):]
    pems = pem_data.split(_BEGIN_CERT_TOKEN)
    output = []
    error = []
    rc = 0
    for pem in pems:
        if pem:
            pem = _BEGIN_CERT_TOKEN + pem
            pem_file = generate_certifcate_file(pem)
            po = Popen([_CMD_OPENSSL, "x509", "-in", pem_file.name],
                        stdin=PIPE, stdout=PIPE, stderr=PIPE)
            out, err = [p.decode() if hasattr(p, 'decode') else p
                        for p in po.communicate()]
            if out:
                output.append(out)
            if err:
                error.append(err)
            if po.returncode:
                rc = po.returncode
            remove_certificate_file(pem_file.name)
    return rc, "".join(output), "".join(error)
