# Copyright (C) 2021 VMware, Inc. All rights reserved.
# -- VMware Confidential
import abc
import logging

from sha.cli.cli_utils import CommandResponse, CLI_RESP_HELP_DESC, CLI_RESP_HELP_OPTIONS, \
    CLI_RESP_HELP_OPTION_NAME, CLI_RESP_HELP_OPTION_VALUE, CLI_RESP_HELP_OPTION_REQUIRED, \
    CLI_RESP_HELP_OPTION_DESC


# CLI commands
CLI_CMD_LIST_CMD = "list_cmd"
CLI_CMD_GET_LOG_LEVEL = "get_log_level"
CLI_CMD_SET_LOG_LEVEL = "set_log_level"


# CLI exception
class CliCommandException(Exception):
    pass


# CLI command providers
class CliCommandProvider(abc.ABC):

    def __init__(self, args):
        self._args = args

    @abc.abstractmethod
    def execute(self):
        pass

    def help(self):
        result = {
            CLI_RESP_HELP_DESC: self.summary(),
            CLI_RESP_HELP_OPTIONS: {}
        }
        return CommandResponse(result, 0, None)

    @classmethod
    @abc.abstractmethod
    def summary(cls):
        pass

    @property
    def args(self):
        return self._args


class CliListCommandProvider(CliCommandProvider):

    def __init__(self, args):
        super(CliListCommandProvider, self).__init__(args)

    def execute(self):
        result = {}
        for cmd, cmd_provider in CLI_CMD_TABLE.items():
            if cmd == CLI_CMD_LIST_CMD:
                continue
            result[cmd] = cmd_provider.summary()
        return CommandResponse(result, 0, None)

    @classmethod
    def summary(cls):
        return "List all the commands"


class CliGetLogLevelProvider(CliCommandProvider):

    def __init__(self, args):
        super(CliGetLogLevelProvider, self).__init__(args)

    def execute(self):
        import nsx_logging
        logger = nsx_logging.getLogger()
        level_map = {
            logging.DEBUG: 'debug',
            logging.INFO: 'info',
            logging.WARNING: 'warn',
            logging.ERROR: 'error',
            logging.CRITICAL: 'fatal'
        }
        level = logger.getEffectiveLevel()
        result = {
            'level': level_map[level]
        }
        return CommandResponse(result, 0, None)

    @classmethod
    def summary(cls):
        return "Get the effective logging level"


class CliSetLogLevelProvider(CliCommandProvider):

    def __init__(self, args):
        super(CliSetLogLevelProvider, self).__init__(args)

    def execute(self):
        import nsx_logging
        logger = nsx_logging.getLogger()
        level_map = {
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warn': logging.WARNING,
            'error': logging.ERROR,
            'fatal': logging.CRITICAL
        }
        level_str = self.args.get('level')
        if level_str is None:
            raise CliCommandException("Required argument 'level' is missing")
        level = level_map.get(level_str)
        if level is None:
            raise CliCommandException("Invalid value {} for argument 'level'".format(level_str))
        logger.setLevel(level)
        result = {
            'level': level_str
        }
        return CommandResponse(result, 0, None)

    def help(self):
        result = {
            CLI_RESP_HELP_DESC: self.summary(),
            CLI_RESP_HELP_OPTIONS: [
                {
                    CLI_RESP_HELP_OPTION_NAME: 'level',
                    CLI_RESP_HELP_OPTION_VALUE: '[debug|info|warn|error|fatal]',
                    CLI_RESP_HELP_OPTION_REQUIRED: True,
                    CLI_RESP_HELP_OPTION_DESC: 'Logging level to be specified'
                }
            ]
        }
        return CommandResponse(result, 0, None)

    @classmethod
    def summary(cls):
        return "Set the logging level"


# CLI command table
CLI_CMD_TABLE = {
    CLI_CMD_LIST_CMD: CliListCommandProvider,
    CLI_CMD_GET_LOG_LEVEL: CliGetLogLevelProvider,
    CLI_CMD_SET_LOG_LEVEL: CliSetLogLevelProvider
}
