# Copyright (C) 2020 - 2021 VMware, Inc. All rights reserved.
# -- VMware Confidential

import os
import json
import traceback
import sha.config
import time
import re

from sha.contrib.metric.edge_base import EdgeBase
from sha.contrib.metric.utils import run_command, timeout_cache
from psutil import net_io_counters, Process, getloadavg, cpu_count, boot_time, swap_memory, users, cpu_times

KB = 1024
ClocksPerSec = float(100)

net_field_keys = ['bytes_sent',
                  'bytes_recv',
                  'packets_sent',
                  'packets_recv',
                  'errin',
                  'errout',
                  'dropin',
                  'dropout']

disk_field_keys = {'block_size': 0,
                   'block': 2,
                   'block_free': 3,
                   'block_avail': 4}

diskio_field_keys = {'io_time': {'index': 12, 'accumulative': True},
                     'iops_in_progress': {'index': 11, 'accumulative': False},
                     'read_time': {'index': 6, 'accumulative': True},
                     'weighted_io_time': {'index': 13, 'accumulative': True},
                     'write_time': {'index': 10, 'accumulative': True}}

proc_state_map = {'W': "paging",
                  'U': "blocked",
                  'D': "blocked",
                  'L': "blocked",
                  'Z': "zombies",
                  'X': "dead",
                  'T': "stopped",
                  't': "stopped",
                  'R': "running",
                  'S': "sleeping",
                  'I': "idle",
                  '?': "unknown"}

swap_field_keys = {'total': 0,
                   'free': 2}

cpu_field_keys = {'user': 1,
                  'nice': 2,
                  'system': 3,
                  'idle': 4,
                  'iowait': 5,
                  'irq': 6,
                  'softirq': 7,
                  'guest': 9}

class SysStats(EdgeBase):
    def __init__(self):
        super().__init__()
        self._logger = sha.config.get_logger(__name__)
        self._diskio_stats = {}
        self._cpu_stats = {}

    def GetProcstatStats(self):
        """ProcStat plugin"""
        """
        Get the specified processes' PID.
        ~# pgrep -f 'nsx.*|kni-single|systemd-journal|lb-dispatcher|telegraf|datapathd|dp-fp.*|tcpdump|iked'
        288
        702
        798
        1150
        1204
        ......
        """
        if sha.config.is_ua():
            service_cmd = ['pgrep', '-f',
                           'opsAgent|async-replicator|syslogd|sshd|ntp|upgrade-agent|nsx-platform-client|appl-proxy']
        else:
            service_cmd = ['pgrep', '-f',
                           'nsx.*|kni-single|systemd-journal|lb-dispatcher|telegraf|datapathd|dp-fp.*|tcpdump|iked|opsAgent']
        try:
            output = run_command(service_cmd, shell=False) or None
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get edge service list: {}, {}".format(e, trace))
            return None

        service_list = output.splitlines()
        """
        Get the specified processes's name, user, memory usage and cpu usage.
        """
        it = iter(service_list)
        procstat_stats_dict = {}
        while True:
            stats = {}
            try:
                pid = next(it)
                self._logger.debug("pid is {}".format(pid))
                path = os.path.join('/proc', pid)
                if not os.path.isdir(path):
                    continue
                p = Process(int(pid))
                with p.oneshot():
                    name = p.name()
                    if "appctl" in name:
                        self._logger.info("Ignore appctl {}, which run shortly, probably quit soon".format(name))
                        continue
                    usr = p.username()
                    cpu_usage = float(p.cpu_percent(0))
                    memory_usage = float(p.memory_percent())
                    stats['cpu_usage'] = cpu_usage
                    stats['memory_usage'] = memory_usage
                    stats['user'] = usr
            except StopIteration as e:
                break;
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning("Failed to get the service info: {}, {}".format(e, trace))
                continue
            procstat_stats_dict[name] = stats

        self._logger.debug("GetProcstatStats:procstat_stats is {}".format(procstat_stats_dict))
        return procstat_stats_dict

    def GetNetStatsPerNic(self):
        """Net per NIC"""
        """
        Get Net Io stats per NIC.
        >>> psutil.net_io_counters()
        snetio(bytes_sent=14508483, bytes_recv=62749361, packets_sent=84311, packets_recv=94888, errin=0, errout=0, dropin=0, dropout=0)
        """
        try:
            output = net_io_counters(pernic=True, nowrap=True)
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get net io stats: {}, {}".format(e, trace))
            return None

        net_stats_dict = {}
        for nic, stats in output.items():
            net_stats = {}
            """Skip loopback interface"""
            if nic == "lo":
                continue
            for key in net_field_keys:
                net_stats[key] = stats._asdict()[key]
            net_stats_dict[nic] = net_stats

        self._logger.debug("GetNetStatsPerNic:net_stats_dict is {}".format(net_stats_dict))
        return net_stats_dict

    def GetMemStats(self):
        """Mem plugin"""
        """
        Get memory information
        ~# cat /proc/meminfo
        MemTotal:        3834444 kB
        MemFree:          275224 kB
        MemAvailable:    2010328 kB
        Buffers:          351720 kB
        Cached:          1209292 kB
        ......
        """
        mem_cmd = ['cat', '/proc/meminfo']
        try:
            output = run_command(mem_cmd, shell=False) or None
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get mem stats: {}, {}".format(e, trace))
            return None

        stats_list = output.splitlines()
        mem_stats_dict = {}
        mem_stats = {}
        for item in stats_list:
           item_list = item.split(':')
           mem_stats[item_list[0]] = item_list[1].rstrip('kB').lstrip(' ').rstrip(' ') if item_list[1] else 0

        mem_stats_dict['active'] = int(mem_stats['Active']) * KB
        mem_stats_dict['available'] = int(mem_stats['MemAvailable']) * KB
        mem_stats_dict['total'] = int(mem_stats['MemTotal']) * KB
        mem_stats_dict['buffered'] = int(mem_stats['Buffers']) * KB
        mem_stats_dict['cached'] = int(mem_stats['Cached']) * KB
        mem_stats_dict['dirty'] = int(mem_stats['Dirty']) * KB
        mem_stats_dict['free'] = int(mem_stats['MemFree']) * KB
        mem_stats_dict['huge_page_size'] = int(mem_stats['Hugepagesize']) * KB
        mem_stats_dict['huge_pages_free'] = int(mem_stats['HugePages_Free']) * KB
        mem_stats_dict['huge_pages_total'] = int(mem_stats['HugePages_Total']) * KB
        mem_stats_dict['inactive'] = int(mem_stats['Inactive']) * KB
        mem_stats_dict['page_tables'] = int(mem_stats['PageTables']) * KB
        mem_stats_dict['swap_cached'] = int(mem_stats['SwapCached']) * KB
        mem_stats_dict['swap_free'] = int(mem_stats['SwapFree']) * KB
        mem_stats_dict['swap_total'] = int(mem_stats['SwapTotal']) * KB

        self._logger.debug("GetMemStats:mem_stats_dict is {}".format(mem_stats_dict))
        return mem_stats_dict

    def GetKernelStats(self):
        """Kernel plugin"""
        """
        Get processes stats
        ~#cat /proc/stat
        ......
        ctxt 1592611744
        btime 1602675885
        processes 1638563
        procs_running 7
        procs_blocked 0
        softirq 92706706 0 33312782 2430451 10946554 3949901 0 67562 0 1 41999455
        """
        kernel_cmd = ['cat', '/proc/stat']
        try:
            output = run_command(kernel_cmd, shell=False) or None
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get kernel stats: {}, {}".format(e, trace))
            return None

        stats_list = output.splitlines()
        kernel_stats_dict = {}
        for item in stats_list:
            stats = [each.strip() for each in item.split(' ') if each]
            if 'btime' in item:
                kernel_stats_dict['boot_time'] = int(stats[1])

        self._logger.debug("GetKernelStats:kernel_stats_dict is {}".format(kernel_stats_dict))
        return kernel_stats_dict

    def GetDiskList(self):
        """
        Get device mapper list
        ~# ls -l /dev/mapper
        total 0
        crw------- 1 root root 10, 236 Oct 14 11:44 control
        lrwxrwxrwx 1 root root       7 Oct 14 11:44 nsx-config -> ../dm-0
        lrwxrwxrwx 1 root root       7 Oct 14 11:44 nsx-config__bak -> ../dm-1
        lrwxrwxrwx 1 root root       7 Oct 14 11:44 nsx-image -> ../dm-3
        lrwxrwxrwx 1 root root       7 Oct 14 11:44 nsx-swap_1 -> ../dm-6
        lrwxrwxrwx 1 root root       7 Oct 14 11:44 nsx-tmp -> ../dm-2
        lrwxrwxrwx 1 root root       7 Oct 14 11:44 nsx-var+dump -> ../dm-5
        lrwxrwxrwx 1 root root       7 Oct 14 11:44 nsx-var+log -> ../dm-4
        """
        dm_cmd = ['ls', '-l', '/dev/mapper']
        try:
            output = run_command(dm_cmd, shell=False) or None
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get dev mapper list: {}, {}".format(e, trace))
            return None

        dm_dict = {}
        dm_list = [each.strip() for each in output.splitlines() if "dm-" in each]
        it = iter(dm_list)
        while True:
            try:
                row = next(it)
                [path, mapper] = row.split('-> ../')
                mapper = mapper.rstrip('\n')
                path = [each.strip() for each in path.split(' ') if each][-1]
                dm_dict[path] = mapper
            except StopIteration as e:
                break
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning("Failed to analyze dev mapper list: {}, {}".format(e, trace))
                continue

        """
        Get mount information
        ~# cat /proc/self/mounts
        ......
        /dev/sda2 / ext4 rw,errors=remount-ro 0 0
        /dev/sda3 /os_bak ext4 rw,nodev,relatime 0 0
        /dev/sda1 /boot ext4 rw,nodev,relatime 0 0
        /dev/mapper/nsx-image /image ext4 rw,nodev,relatime 0 0
        /dev/mapper/nsx-config /config ext4 rw,nodev,relatime 0 0
        /dev/mapper/nsx-var+dump /var/dump ext4 rw,nodev,relatime 0 0
        /dev/mapper/nsx-config__bak /config_bak ext4 rw,nodev,relatime 0 0
        /dev/mapper/nsx-tmp /tmp ext4 rw,nosuid,nodev,noexec,relatime 0 0
        /dev/mapper/nsx-var+log /var/log ext4 rw,nodev,relatime 0 0
        ......
        """
        dm_list = output.splitlines()
        disk_cmd = ['cat', '/proc/self/mounts']
        try:
            output = run_command(disk_cmd, shell=False) or None
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get disk stats: {}, {}".format(e, trace))
            return None

        disk_list = output.splitlines()
        disk_dict = {}
        it = iter(disk_list)
        while True:
            try:
                item = next(it)
                item_list = [each.strip() for each in item.split(' ') if each]
                [device, path, fs, mode, dummy1, dummy2] = item_list
                """ Skip /os_bak and /config_bak """
                if path.endswith("bak"):
                    continue
                mode = mode.split(',')[0].strip()
                if (fs == 'ext4') and (mode == 'rw'):
                    if '/dev/mapper' in device:
                        device = dm_dict[device.lstrip('/dev/mapper')]
                    else:
                        device = device.lstrip('/dev/')
                    disk_dict[path] = device
            except StopIteration as e:
                break;
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning("Failed to get the disk info: {}, {}".format(e, trace))
                continue

        return disk_dict

    def GetDiskStats(self):
        """Disk plugin"""
        """
        Get disk list
        """
        disk_dict = self.GetDiskList()
        self._logger.debug("GetDiskStats:disk_dict is {}".format(disk_dict))
        if not disk_dict:
            return None

        """
        Get disk stats
        >>> os.statvfs('/')
        posix.statvfs_result(f_bsize=4096, f_frsize=4096, f_blocks=4888635, f_bfree=3935414, f_bavail=3681322, f_files=1250928, f_ffree=1161261, f_favail=1161261, f_flag=0, f_namemax=255)
        """
        disk_stats_dict = {}
        """key is the path, val is the device name"""
        for path,device in disk_dict.items():
            stats = os.statvfs(path)
            disk_stats = {'free': int(stats[disk_field_keys['block_avail']]) * int(stats[disk_field_keys['block_size']])}
            disk_stats['total'] = int(stats[disk_field_keys['block']]) * int(stats[disk_field_keys['block_size']])
            disk_stats['used'] = int(disk_stats['total']) - int(stats[disk_field_keys['block_free']]) * int(stats[disk_field_keys['block_size']])
            disk_stats['device'] = device
            disk_stats_dict[path] = disk_stats

        self._logger.debug("GetDiskStats:disk_stats_dict is {}".format(disk_stats_dict))
        return disk_stats_dict

    def GetDiskioStats(self):
        """Diskio plugin"""
        """
        Get disk list
        """
        disk_dict = self.GetDiskList()
        self._logger.debug("GetDiskioStats:disk_dict is {}".format(disk_dict))
        if not disk_dict:
            return None

        """
        Get disk stats
        ~# cat /proc/diskstats
          8       0 sda 34353 2865 2063454 131494 9381044 5694742 369548960 171265057 0 5393144 170541088 0 0 0 0
          8       1 sda1 143 0 8866 423 28 21 392 776 0 896 1156 0 0 0 0
          8       2 sda2 30695 2840 1709034 97546 854530 3817107 41132864 33257225 0 1055240 33118884 0 0 0 0
          8       3 sda3 107 0 8570 209 13 3 128 1086 0 1284 1312 0 0 0 0
          8       4 sda4 2 0 4 1 0 0 0 0 0 4 4 0 0 0 0
          8       5 sda5 3289 25 330788 33199 8526473 1877611 328415576 138005970 0 4886256 137414668 0 0 0 0
        252       0 dm-0 202 0 6898 2184 17362 0 152704 96544 0 77420 98724 0 0 0 0
        252       1 dm-1 153 0 6506 1608 16 0 128 452 0 1280 2060 0 0 0 0
        252       2 dm-2 185 0 6738 1872 116767 0 1304256 660404 0 193220 662276 0 0 0 0
        252       3 dm-3 154 0 6514 856 16 0 128 408 0 696 1264 0 0 0 0
        252       4 dm-4 2298 0 292722 26824 10269915 0 326958232 147982104 0 4784180 148052004 0 0 0 0
        252       5 dm-5 157 0 6634 1288 16 0 128 392 0 1040 1680 0 0 0 0
        252       6 dm-6 98 0 4240 148 0 0 0 0 0 72 148 0 0 0 0
        """
        diskio_cmd = ['cat', '/proc/diskstats']
        try:
            output = run_command(diskio_cmd, shell=False) or None
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get diskio stats: {}, {}".format(e, trace))
            return None

        diskio_stats_dict = {}
        it = iter(output.splitlines())
        while True:
            try:
                item = next(it)
                stats = [each.strip() for each in item.split(' ') if each]
                name = stats[2]
                """Only partition with fs as ext4 and mode as rw would be considered for collection"""
                if name not in disk_dict.values():
                    continue
                path = list(disk_dict.keys())[list(disk_dict.values()).index(name)]
                diskio_stats = {}
                if not self._diskio_stats.get(path, None):
                    self._diskio_stats[path] = {}
                for key, detail in diskio_field_keys.items():
                    if key == 'read_bytes' or key == 'write_bytes':
                        factor = 512
                    else:
                        factor = 1
                    value = int(stats[detail['index']]) * factor
                    if detail['accumulative'] == True:
                        origin_value = self._diskio_stats[path].get(key, None)
                        if origin_value and value >= origin_value:
                            diskio_stats[key] = value - origin_value
                        else:
                            diskio_stats[key] = 0
                        self._diskio_stats[path][key] = value
                    else:
                        diskio_stats[key] = value
                    diskio_stats['device'] = name
            except StopIteration as e:
                break;
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning("Failed to get the diskio info: {}, {}".format(e, trace))
                continue
            diskio_stats_dict[path] = diskio_stats

        self._logger.debug("GetDiskioStats:diskio_stats_dict is {}".format(diskio_stats_dict))
        return diskio_stats_dict

    def GetProcessesStats(self):
        """Processes plugin"""
        """
        Get processes stats
        ~# cat /proc/1/status
        Name:	systemd
        Umask:	0000
        State:	S (sleeping)
        Threads:	1
        ......
        """
        proc_stats_dict = {
            'running': 0,
            'sleeping': 0,
            'blocked': 0,
            'dead': 0,
            'idle': 0,
            'paging': 0,
            'stopped': 0,
            'unknown': 0,
            'zombies': 0,
            'total': 0,
            'total_threads': 0
        }
        for item in os.listdir('/proc'):
            if not item.isdigit():
                continue
            path = os.path.join('/proc', item)
            if not os.path.isdir(path):
                continue
            path = os.path.join(path, 'status')
            state_name = None
            try:
                with open(path, 'r') as f:
                    for line in f.readlines():
                        if 'State' in line:
                            state = line[line.find(':') + 2 : line.rfind('(')].strip(" ")
                            self._logger.debug("GetProcessesStats: process {} status is {}".format(item, state))
                            state_name = proc_state_map.get(state, None)
                            if not state_name:
                                self._logger.warning("Unknown state {}".format(state))
                        if 'Threads' in line:
                            threads = int(line.split(':\t')[1].strip())
                    if state_name and proc_stats_dict.get(state_name, None) is not None:
                        proc_stats_dict[state_name] += 1
                    proc_stats_dict['total'] += 1
                    proc_stats_dict['total_threads'] += threads
            except Exception as e:
                trace = traceback.format_exc()
                self._logger.warning("Failed to get info from {}: {}, {}".format(path, e, trace))
                continue

        self._logger.debug("GetProcessesStats:proc_stats_dict is {}".format(proc_stats_dict))
        return proc_stats_dict

    def GetSystemStats(self):
        """System plugin"""
        """
        Get system stats
        >>> psutil.getloadavg()
        (3.14, 3.89, 4.67)
        """
        sys_stats_dict = {}
        try:
            output = getloadavg()
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get average load: {}, {}".format(e, trace))
            output = None

        if output:
            sys_stats_dict['load1'] = float(output[0])

        """
        >>> psutil.cpu_count()
        10
        """
        try:
            output = cpu_count(logical=True)
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get cpu count: {}, {}".format(e, trace))
            output = None

        if output:
            sys_stats_dict['n_cpus'] = output

        """
        >>> psutil.boot_time()
        1389563460.0
        """
        try:
            output = boot_time()
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get boot time: {}, {}".format(e, trace))
            output = None

        if output:
            sys_stats_dict['uptime'] = int(output)

        self._logger.debug("GetSystemStats:sys_stats_dict is {}".format(sys_stats_dict))
        return sys_stats_dict

    def GetSwapStats(self):
        """Swap plugin"""
        """
        Get swap memory stats
        >>> psutil.swap_memory()
        sswap(total=2097147904L, used=886620160L, free=1210527744L, percent=42.3, sin=1050411008, sout=1906720768)
        """
        try:
            output = swap_memory()
        except Exception as e:
            trace = traceback.format_exc()
            self._logger.warning("Failed to get swap stats: {}, {}".format(e, trace))
            return None

        swap_stats_dict = {}
        swap_stats_dict['total'] = int(output[swap_field_keys['total']])
        swap_stats_dict['free'] = int(output[swap_field_keys['free']])
        swap_stats_dict['used'] = int(swap_stats_dict['total'] - swap_stats_dict['free'])

        self._logger.debug("GetSwapStats:swap_stats_dict is {}".format(swap_stats_dict))
        return swap_stats_dict

    def GetCpuStats(self):
        """Cpu plugin"""
        """
        Get cpu stats
        ~#cat /proc/stat
        cpu  2043380 94503 2007975 43381436 511374 0 15737 0 0 0
        cpu0 518561 0 581235 22633725 266092 0 5916 0 0 0
        cpu1 1524818 94503 1426740 20747710 245282 0 9820 0 0 0
        """
        if not sha.config.is_ua():
            core_dict = {}
            with open('/config/vmware/edge/config.json', 'r') as f:
                lines = [line.strip('"').strip(' ').strip('\n') for line in f.readlines() if 'list' in line]
                for line in lines:
                    items = line.split(':')
                    cores = [core.strip(' ').strip(',').strip('"') for core in items[1].split(',') if core]
                    core_dict[items[0].strip('"')] = cores
            core_dict['dp_corelist'] = []
            for cpu in core_dict['corelist']:
                if cpu not in core_dict['service_corelist']:
                    core_dict['dp_corelist'].append(cpu)
            self._logger.debug("GetCpuStats:core_dict is {}".format(core_dict))

        cpu_stats_dict = {}
        with open('/proc/stat', 'r') as f:
            for line in f.readlines():
                if 'cpu' not in line:
                    continue
                stats = [each.strip() for each in line.split(' ') if each]
                name = stats[0]

                cpu_idx = None
                cpu_stats = {}
                cpu_stats_usage = {}
                match = re.search(r'cpu(\d+)', name)
                if match:
                    cpu_idx = match.group(1)
                else:
                    continue
                if cpu_idx:
                    if not sha.config.is_ua():
                        if cpu_idx in core_dict['local_cpulist'] and cpu_idx in core_dict['dp_corelist']:
                            '''Ship the collection for dp cores'''
                            continue
                        else:
                            cpu_stats_usage['tag'] = 'service-core'

                user_usage = None
                guest_usage = None
                if not self._cpu_stats.get(name, None):
                    self._cpu_stats[name] = {}
                for key, index in cpu_field_keys.items():
                    origin_value = self._cpu_stats[name].get(key, None)
                    value = float(stats[index]) / ClocksPerSec
                    import math
                    if origin_value != None and \
                       (value > origin_value or math.isclose(value, origin_value) == True):
                        cpu_stats_usage[key] = value
                        if key == 'user':
                            if guest_usage != None:
                                cpu_stats_usage['user'] = cpu_stats_usage['user'] - guest_usage - \
                                                          self._cpu_stats[name]['user']
                                self._cpu_stats[name]['user'] = value
                            else:
                                user_usage = value
                        else:
                            if key == 'guest':
                                if user_usage != None:
                                    cpu_stats_usage['user'] = user_usage - cpu_stats_usage['guest'] - \
                                                              self._cpu_stats[name]['user']
                                    self._cpu_stats[name]['user'] = user_usage
                                else:
                                    guest_usage = value
                            cpu_stats_usage[key] = cpu_stats_usage[key] - self._cpu_stats[name][key]
                            self._cpu_stats[name][key] = value
                    else:
                        cpu_stats_usage[key] = 0
                        self._cpu_stats[name][key] = value

                self._logger.debug("GetCpuStats:name is {}".format(name))
                cpu_stats_dict[name] = cpu_stats_usage

        self._logger.debug("GetCpuStats:cpu_stats_dict is {}".format(cpu_stats_dict))
        return cpu_stats_dict

    @timeout_cache(2)
    def usage(self, *args, **kwargs):
        procstat_stats = self.GetProcstatStats()
        net_stats = self.GetNetStatsPerNic()
        mem_stats = self.GetMemStats()
        kernel_stats = self.GetKernelStats()
        disk_stats = self.GetDiskStats()
        diskio_stats = self.GetDiskioStats()
        proc_stats = self.GetProcessesStats()
        sys_stats = self.GetSystemStats()
        swap_stats = self.GetSwapStats()
        cpu_stats = self.GetCpuStats()
        return {'procstat': procstat_stats, \
                'net': net_stats, \
                'mem': mem_stats, \
                'kernel': kernel_stats, \
                'disk': disk_stats, \
                'diskio': diskio_stats, \
                'processes': proc_stats, \
                'system': sys_stats, \
                'swap': swap_stats, \
                'cpu': cpu_stats}, sha.config.NODE_TYPE
