# Copyright 2013-2017 VMware, Inc.  All rights reserved.
# -*- mode: Python; coding: utf-8 -*-
# **********************************************************
#
"""
Implementation of appliance SNMP API.
"""

__author__ = "VMware, Inc."

import shutil
import logging
import os
import json
import subprocess as sb
from pkg_defs import *

# from vmw_snmp import null_dev

DEFAULT_SNMP_PORT = "161"
DEF_TRAP_PORT = "162"
null_dev = open(os.devnull, "w")
logger = logging.getLogger(__name__)

# where to write json file for agent
dest_path = '/etc/vmware/appliance/firewall/vmware-snmp'
tmp_dest_path = '/tmp/vmware-snmp-fw-tmp'

fw_pgm_path = '/usr/lib/applmgmt/networking/bin/firewall-reload'
errlog_path = "/var/log/vmw_snmp_firewall.log"

def have_firewall():
   return os.path.exists(fw_pgm_path)

def read_log(path):
    ''' read contents of path and return as string'''
    try:
        with open(path, 'r') as errlog:
            return errlog.read()
    except IOError:
        return "n/a"

def _reload_fw():
    '''force fw to reload config'''
    app_args = [fw_pgm_path]
    try:
       with open(errlog_path, 'w') as errlog:
           logger.debug("_reload_fw: invoking command: %s" % " ".join(app_args))
           sb.check_call(app_args, stdin=null_dev, stderr=errlog, stdout=errlog)
    except Exception as err:
        logmsg = read_log(errlog_path)
        msg = "_reload_fw: command '%s' failed {%s}, stdout={%s}" % (" ".join(app_args), err, logmsg)
        logger.error(msg)
        raise Exception(msg)

def _generate_fw_config(cfg):
    '''convert data from cfg into firewall ruleset in json format '''
    try:
        with open(tmp_dest_path, "w") as cfgFile:
            json.dump(cfg, cfgFile)
        msg = "_generate_fw_config: renaming config file %s to %s" % (tmp_dest_path, dest_path)
        logger.debug(msg)
        shutil.move(tmp_dest_path, dest_path)
        logger.info("_generate_fw_config: %s updated." % dest_path)
    except IOError as err:
        msg = '_generate_fw_config: move %s to %s failed %s' % \
                 (tmp_dest_path, dest_path, str(err))
        logging.error(msg)
        raise Exception(msg)

def _get_v1_port(v1tgt):
    '''return string of port number for this rule'''
    def_port = str(DEF_TRAP_PORT)
    rslt = v1trp_re.match(v1tgt)
    if rslt is None:
        logger.error("_get_v1_port: regular expression failed for input '%s', ignoring" % v1tgt)
        return def_port
    if rslt.group('port'):
        return rslt.group('port')
    else:
        return def_port

def _get_v3_port(v3tgt):
    '''return string of port number for this rule'''
    def_port = str(DEF_TRAP_PORT)
    rslt = v3target_re.match(v3tgt)
    if rslt is None:
        logger.error("_get_v3_port: regular expression failed for input '%s', ignoring" % v3tgt)
        return def_port
    if rslt.group('port'):
        return rslt.group('port')
    else:
        return def_port

def _add_outbound_ports(kvp, rs):
    '''add any outbound ports found in the config file to rs list '''
    port_set = set()
    v1tgts = kvp.get(TARGETS, [])
    for item in v1tgts:
        if len(item) > 0:
            port_set.add(_get_v1_port(item))
    v3tgts = kvp.get(V3TARGETS, [])
    for item in v3tgts:
        if len(item) > 0:
            port_set.add(_get_v3_port(item))
    for ps in port_set:
        rule = { "direction" : "outbound",
                 "protocol"  : "udp",
                 "porttype"  : "dst",
                 "port"      : ps,
                 "portoffset": 0
        }
        rs.append(rule)

def _gen_ruleset(kvp):
    '''given snmp config, return cloudvm ruleset'''
    state = False
    port_num = str(kvp.get(PORT, DEFAULT_SNMP_PORT))
    poll_port = { "direction" : "inbound",
               "protocol"  : "udp",
               "porttype"  : "dst",
               "port"      : str(port_num),
               "portoffset": 0
    }
    poll_port["port"] = port_num
    if kvp.get(ENABLE, "false") == "true":
        state = True
    rs = [poll_port]
    _add_outbound_ports(kvp, rs)
    fw = { "enable" : state, "rules" : rs}
    fw_cfg = { 'firewall' : fw }
    return fw_cfg

def update_firewall(cfg):
    '''cfg is the dict of /etc/vmware/snmp/snmp.xml '''
    if not have_firewall():
       return
    logger.info("update_firewall: updating ruleset started")
    ruleset = _gen_ruleset(cfg)
    _generate_fw_config(ruleset)
    _reload_fw()
    logger.info("update_firewall: updating ruleset completed")


if __name__ == '__main__':
    print("testing running update_firewall started")
    cfg = { ENABLE : "true", PORT : "193",
            TARGETS : [ '192.160.0.1@1162/mycom1',
                        '192.168.1.1/mycom2',
                        '10.20.102.103@65534/mycom3']
    }
    update_firewall(cfg)
    print("running update_firewall completed")
